/* Assembler: scanner. */

%{
  #include <config.h>
  #include <stdio.h>
  #include <limits.h>
  #include <easy-machine/easy-machine-utility.h>
  #include <easy-machine/easy-machine.h>
  #include <easy-machine/easy-machine-machine.h>
  #include <easy-machine/easy-machine-instructions-machine-generated.h>

  #include <easy-machine/easy-machine-assembler.h>
  #include <easy-machine/easy-machine-asm-parser.h>
  //#include <easy-machine/easy-machine-asm-scanner.h>

  /* Re-declare two automatically-defined flex functions with
     __attribute__((unused)), to prevent an annoying GCC warning. */
  static int input  (yyscan_t yyscanner)
    __attribute__ ((unused));
  static void yyunput (int c, char * yy_bp , yyscan_t yyscanner)
    __attribute__ ((unused));
%}
%option bison-bridge bison-locations nodefault noyywrap prefix="easy_machine_"
%option reentrant yylineno

BINARY_NATURAL       0b[0-1]+
OCTAL_NATURAL        0(o?)[0-7]+
DECIMAL_NATURAL      (0|([1-9][0-9]*))
HEXADECIMAL_NATURAL  0x([0-9]|[a-f]|[A-F])+

BINARY_INTEGER       [-+]?{BINARY_NATURAL}
OCTAL_INTEGER        [-+]?{OCTAL_NATURAL}
DECIMAL_INTEGER      [-+]?{DECIMAL_NATURAL}
HEXADECIMAL_INTEGER  [-+]?{HEXADECIMAL_NATURAL}

STRING_LITERAL   \"([^\"\\]|\\\"|\\\\)*\"

IDENTIFIER       [_a-zA-Z][-+._~@/\\a-zA-Z0-9]*
LABEL_PREFIX     "$"
REGISTER         "%r"{DECIMAL_NATURAL}

WHITESPACE       [\ \t\n\r]+
COMMENT          "#".*(\n|\r)

%%

({WHITESPACE}|{COMMENT})+ { /* Do nothing. */ }
{BINARY_INTEGER}          { return SIGNED_BINARY_LITERAL; }
{OCTAL_INTEGER}           { return SIGNED_OCTAL_LITERAL; }
{DECIMAL_INTEGER}         { return SIGNED_DECIMAL_LITERAL; }
{HEXADECIMAL_INTEGER}     { return SIGNED_HEXADECIMAL_LITERAL; }
{BINARY_INTEGER}[uU]      { return UNSIGNED_BINARY_LITERAL; }
{OCTAL_INTEGER}[uU]       { return UNSIGNED_OCTAL_LITERAL; }
{DECIMAL_INTEGER}[uU]     { return UNSIGNED_DECIMAL_LITERAL; }
{HEXADECIMAL_INTEGER}[uU] { return UNSIGNED_HEXADECIMAL_LITERAL; }
{REGISTER}                { return REGISTER; }
"+"                       { return PLUS; }
"-"                       { return MINUS; }
"*"                       { return TIMES; }
"/"                       { return DIV; }
"%"                       { return MOD; }
"+"[uU]                   { return UNSIGNED_PLUS; }
"-"[uU]                   { return UNSIGNED_MINUS; }
"*"[uU]                   { return UNSIGNED_TIMES; }
"/"[uU]                   { return UNSIGNED_DIV; }
"%"[uU]                   { return UNSIGNED_MOD; }
"("                       { return OPEN_PARENS; }
")"                       { return CLOSE_PARENS; }
".code"                   { return DOT_CODE; }
".readonly"               { return DOT_READONLY; }
".ro"                     { return DOT_READONLY; }
".readwrite"              { return DOT_READWRITE; }
".rw"                     { return DOT_READWRITE; }
".byte"                   { return DOT_BYTE; }
".word"                   { return DOT_WORD; }
".stringz"                { return DOT_STRINGZ; }
".string"                 { return DOT_STRING; }
".space"                  { return DOT_SPACE; }
{STRING_LITERAL}          { return STRING_LITERAL; }
{IDENTIFIER}              { return INSTRUCTION_NAME; }
{LABEL_PREFIX}{IDENTIFIER}    { return LABEL_LITERAL; }
{LABEL_PREFIX}{IDENTIFIER}":" { return LABEL; }
,                         { return COMMA; }
.                         { easy_machine_scan_error (yyscanner); }

%%
