/* Assembler: parser. */

/* This code is included in the generated .c file, but not in the header. */
%{
  #include <config.h>
  #include <stdio.h>
  #include <limits.h>
  #include <easy-machine/easy-machine.h>
  #include <easy-machine/easy-machine-machine.h>
  #include <easy-machine/easy-machine-utility.h>
  #include <easy-machine/easy-machine-machine.h>
  #include <easy-machine/easy-machine-instructions-machine-generated.h>

  #include <easy-machine/easy-machine-assembler.h>
  #include <easy-machine/easy-machine-asm-parser.h>
  #include <easy-machine/easy-machine-asm-scanner.h>

  /* This is currently a fatal error.  I could longjmp away instead. */
  static void
  easy_machine_error (YYLTYPE *locp,
                struct parser_arg *parser_arg, /*struct easy_machine_routine *routine,*/
                yyscan_t scanner, const char *message)
    __attribute__ ((noreturn));

  /* Just a forward-declaration.  FIXME: the error-reporting code in the easy_machinec
     parser is derived from this, but more advanced; I should integrate it. */
  static void
  easy_machine_simple_error (void *easy_machine_scanner, const char *message)
    __attribute__ ((noreturn));

/* Set the given lvalue, with the %type whose definition is union
   easy_machine_literal, converting easy_machine_get_text (easy_machine_scanner) using the given
   function.  These are useful to convert to a specified base, in the case of
   signed and unsigned literals.  The lvalues will be either
   $$.fixnum or $$.ufixnum , and the functions to
   call will be easy_machine_strtoll or easy_machine_strtoull . */
#define EASY_MACHINE_SET_BINARY(lvalue, function)           \
  do                                                  \
    {                                                 \
      char *text = easy_machine_get_text (easy_machine_scanner);  \
      * strchr (text, 'b') = '0';                     \
      lvalue = function (text, NULL, 2);              \
    }                                                 \
  while (false)
#define EASY_MACHINE_SET_OCTAL(lvalue, function)            \
  do                                                  \
    {                                                 \
      char *text = easy_machine_get_text (easy_machine_scanner);  \
      char *oindex = strchr (text, 'o');              \
      if (oindex != NULL)                             \
        * oindex = '0';                               \
      lvalue = function (text, NULL, 8);              \
    }                                                 \
  while (false)
#define EASY_MACHINE_SET_DECIMAL(lvalue, function)          \
  do                                                  \
    {                                                 \
      char *text = easy_machine_get_text (easy_machine_scanner);  \
      lvalue = function (text, NULL, 10);             \
    }                                                 \
  while (false)
#define EASY_MACHINE_SET_HEXADECIMAL(lvalue, function)      \
  do                                                  \
    {                                                 \
      char *text = easy_machine_get_text (easy_machine_scanner);  \
      lvalue = function (text, NULL, 16);             \
    }                                                 \
  while (false)

#define EASY_MACHINE_SET_OPERATION(result, operand1, operator, operand2)  \
  do                                                                \
    {                                                               \
      (result) = (operand1) operator (operand2);                    \
    }                                                               \
  while (false)

#define EASY_MACHINE_SET_LOCATION  \
  do \
    { \
      parser_arg->routine->line_no \
        = easy_machine_get_lineno (easy_machine_scanner); \
    } \
  while (false)
%}

/* We need a recent enough version of GNU Bison. */
%require "2.3b" /* This is the first version supporting %define api.pure . */

/* Use a prefix different from the default "yy" for the API. */
%name-prefix "easy_machine_"

/* Generate a header file. */
%defines

/* This is a reentrant parser. */
%define api.pure

/* We need to receive location information from the scanner, Bison-style. */
%locations

/* The parser and scanner functions both have additional parameters. */
%lex-param { easy_machine_scan_t easy_machine_scanner }
%parse-param { struct parser_arg *parser_arg }
%parse-param { void* easy_machine_scanner }

/* We don't need a %initial-action block, because the parser receives an already
   initialized routine; see the definition of easy_machine_parse_file_star . */

/* This goes to the header file. */
%code requires {
# include <config.h>

# include <stdio.h>
# include <limits.h>

# include <easy-machine/easy-machine.h>
# include <easy-machine/easy-machine-utility.h>
# include <easy-machine/easy-machine-machine.h>
# include <easy-machine/easy-machine-instructions-machine-generated.h>
# include <easy-machine/easy-machine-assembler.h>

  /* The structure whose pointer is passed to the parser function.  FIXME:
     revert to having just the routine as a parser argument: the VM is reachable
     from the routine. */
  struct parser_arg
  {
    /* The routine to be parsed, allowed but not required to be empty on
       input. */
    struct easy_machine_routine *routine;
  };

  /* Simplified error-reporting facilities calling vmprefix_error, suitable to be
     called from the scanner and the parser without the complicated and
     irrelevant parameters needed by vmprefix_error . */
  void
  easy_machine_scan_error (void *easy_machine_scanner) __attribute__ ((noreturn));
  void
  easy_machine_parse_error (void *easy_machine_scanner) __attribute__ ((noreturn));

  /* Parse a routine for the pointed VM from a file or a string in memory, adding
     code to the pointed VM routine.
     These functions work of course on any VM, but are slightly inconvenient for
     the user to call directly.  For this reason they are wrapped in the vm1.c
     template into VM-specific functions not requiring a VM struct pointer. */
  void
  easy_machine_parse_file_star (FILE *input_file, struct easy_machine_routine *p)
    __attribute__ ((nonnull (1, 2)));
  void
  easy_machine_parse_file (const char *input_file_name, struct easy_machine_routine *p)
    __attribute__ ((nonnull (1, 2)));
  void
  easy_machine_parse_string (const char *string, struct easy_machine_routine *p)
    __attribute__ ((nonnull (1, 2)));
}

%union {
  easy_machine_word literal;
}

%token INSTRUCTION_NAME REGISTER LABEL_LITERAL LABEL COMMA

%token OPEN_PARENS CLOSE_PARENS
%token SIGNED_BINARY_LITERAL SIGNED_OCTAL_LITERAL SIGNED_DECIMAL_LITERAL
%token SIGNED_HEXADECIMAL_LITERAL
%token UNSIGNED_BINARY_LITERAL UNSIGNED_OCTAL_LITERAL UNSIGNED_DECIMAL_LITERAL
%token UNSIGNED_HEXADECIMAL_LITERAL
%token PLUS MINUS TIMES DIV MOD
%token UNSIGNED_PLUS UNSIGNED_MINUS UNSIGNED_TIMES UNSIGNED_DIV UNSIGNED_MOD
%token DOT_BYTE DOT_WORD DOT_STRING DOT_STRINGZ DOT_SPACE
%token STRING_LITERAL
%token DOT_CODE DOT_READONLY DOT_READWRITE
%expect 100 /* FIXME: handle precedence and associativity.  Those shift/reduce
               conflicts should only come from expressions, unless something
               quite big escaped my attention. */
%type <literal> int_expression

%%

routine :
  /* nothing */
| routine change_section
| routine instruction
| routine data_definition
| routine label
;

change_section :
  DOT_CODE
    { EASY_MACHINE_SET_LOCATION;
      easy_machine_asm_enter_section (parser_arg->routine,
                                      easy_machine_asm_section_code); }
| DOT_READONLY
    { EASY_MACHINE_SET_LOCATION;
      easy_machine_asm_enter_section (parser_arg->routine,
                                      easy_machine_asm_section_readonly); }
| DOT_READWRITE
    { EASY_MACHINE_SET_LOCATION;
      easy_machine_asm_enter_section (parser_arg->routine,
                                      easy_machine_asm_section_readwrite); }
;

instruction :
  instruction_name arguments
;

data_definition :
  DOT_BYTE
    { EASY_MACHINE_SET_LOCATION;
      easy_machine_asm_append_bytes (parser_arg->routine); }
  data
| DOT_WORD
    { EASY_MACHINE_SET_LOCATION;
      easy_machine_asm_append_words (parser_arg->routine); }
  data
| DOT_STRING
    { EASY_MACHINE_SET_LOCATION;
      easy_machine_asm_append_bytes (parser_arg->routine); }
  string_literal
| DOT_STRINGZ
    { EASY_MACHINE_SET_LOCATION;
      easy_machine_asm_append_bytes (parser_arg->routine); }
  string_literal
    { easy_machine_asm_append_datum_literal (parser_arg->routine, 0); }
| DOT_SPACE
    { EASY_MACHINE_SET_LOCATION;
      easy_machine_asm_append_bytes (parser_arg->routine); }
  int_expression
    { int byte_no = $3;
      int i;
      for (i = 0; i < byte_no; i ++)
        easy_machine_asm_append_datum_literal (parser_arg->routine, 0); }
;

string_literal:
  STRING_LITERAL
    {
      EASY_MACHINE_SET_LOCATION;
      unsigned char *s
        = (unsigned char *) easy_machine_get_text (easy_machine_scanner);
      for (s ++; * s != '\"'; s ++)
        {
          if (* s == '\\')
            s ++;
          easy_machine_asm_append_datum_literal (parser_arg->routine, * s);
        }
    }
  ;

datum :
  int_expression
    { easy_machine_asm_append_datum_literal (parser_arg->routine, $1); }
| label_datum
  ;

data :
  datum after_datum
  ;

after_datum :
  /* nothing */
| COMMA
    {EASY_MACHINE_SET_LOCATION;}
  datum after_datum
;

label_datum:
  LABEL_LITERAL
    { EASY_MACHINE_SET_LOCATION;
      char *label = easy_machine_get_text (easy_machine_scanner);
      label [strlen (label) - 1] = '\0';  /* Remove the trailing colon. */
      /* Add one to skip the prefix. */
      easy_machine_asm_append_datum_symbolic_label (parser_arg->routine, label + 1);
    }
  ;

label :
  LABEL { EASY_MACHINE_SET_LOCATION;
          char *label = easy_machine_get_text (easy_machine_scanner);
          label [strlen (label) - 1] = '\0';  /* Remove the trailing colon. */
          /* Add one to skip the prefix. */
          easy_machine_asm_append_symbolic_label (parser_arg->routine, label + 1); }
;

instruction_name :
  INSTRUCTION_NAME { EASY_MACHINE_SET_LOCATION;
                     char *name = easy_machine_get_text (easy_machine_scanner);
                     easy_machine_asm_append_instruction_name (parser_arg->routine,
                                                     name); }
;

arguments :
  /* nothing */
| argument after_one_argument
;

after_one_argument :
  /* nothing*/
| COMMA argument after_one_argument
;

int_expression :
  SIGNED_BINARY_LITERAL       { EASY_MACHINE_SET_LOCATION;
                                EASY_MACHINE_SET_BINARY($$,
                                                  strtoll); }
| SIGNED_OCTAL_LITERAL        { EASY_MACHINE_SET_LOCATION;
                                EASY_MACHINE_SET_OCTAL($$,
                                                 strtoll); }
| SIGNED_DECIMAL_LITERAL      { EASY_MACHINE_SET_LOCATION;
                                EASY_MACHINE_SET_DECIMAL($$,
                                                   strtoll); }
| SIGNED_HEXADECIMAL_LITERAL  { EASY_MACHINE_SET_LOCATION;
                                EASY_MACHINE_SET_HEXADECIMAL($$,
                                                       strtoll); }
| UNSIGNED_BINARY_LITERAL     { EASY_MACHINE_SET_LOCATION;
                                EASY_MACHINE_SET_BINARY($$,
                                                  strtoull); }
| UNSIGNED_OCTAL_LITERAL      { EASY_MACHINE_SET_LOCATION;
                                EASY_MACHINE_SET_OCTAL($$,
                                                 strtoull); }
| UNSIGNED_DECIMAL_LITERAL    { EASY_MACHINE_SET_LOCATION;
                                EASY_MACHINE_SET_DECIMAL($$,
                                                   strtoull); }
| UNSIGNED_HEXADECIMAL_LITERAL{ EASY_MACHINE_SET_LOCATION;
                                EASY_MACHINE_SET_HEXADECIMAL($$,
                                                       strtoull); }
| OPEN_PARENS int_expression CLOSE_PARENS { EASY_MACHINE_SET_LOCATION;
                                            $$ = $2; }
| int_expression PLUS int_expression
     { EASY_MACHINE_SET_LOCATION;
       EASY_MACHINE_SET_OPERATION($$, $1, +, $3); }
| int_expression MINUS int_expression
     { EASY_MACHINE_SET_LOCATION;
       EASY_MACHINE_SET_OPERATION($$, $1, -, $3); }
| int_expression TIMES int_expression
     { EASY_MACHINE_SET_LOCATION;
       EASY_MACHINE_SET_OPERATION($$, $1, *, $3); }
| int_expression DIV int_expression
     { EASY_MACHINE_SET_LOCATION;
       EASY_MACHINE_SET_OPERATION($$, $1, /, $3); }
| int_expression MOD int_expression
     { EASY_MACHINE_SET_LOCATION;
       EASY_MACHINE_SET_OPERATION($$, $1, %, $3); }
| int_expression UNSIGNED_PLUS int_expression
     { EASY_MACHINE_SET_LOCATION;
       EASY_MACHINE_SET_OPERATION($$, $1, +, $3); }
| int_expression UNSIGNED_MINUS int_expression
     { EASY_MACHINE_SET_LOCATION;
       EASY_MACHINE_SET_OPERATION($$, $1, -, $3); }
| int_expression UNSIGNED_TIMES int_expression
     { EASY_MACHINE_SET_LOCATION;
       EASY_MACHINE_SET_OPERATION($$, $1, *, $3); }
| int_expression UNSIGNED_DIV int_expression
     { EASY_MACHINE_SET_LOCATION;
       EASY_MACHINE_SET_OPERATION($$, $1, /, $3); }
| int_expression UNSIGNED_MOD int_expression
     { EASY_MACHINE_SET_LOCATION;
       EASY_MACHINE_SET_OPERATION($$, $1, %, $3); }
;

argument :
  int_expression { EASY_MACHINE_SET_LOCATION;
                   easy_machine_asm_append_literal_parameter (parser_arg->routine,
                                                    $1); }
| LABEL_LITERAL  { EASY_MACHINE_SET_LOCATION;
                   char *text = easy_machine_get_text (easy_machine_scanner) + 1; /* Skip the prefix. */
                   easy_machine_asm_append_symbolic_label_parameter (parser_arg->routine,
                                                           text); }
| REGISTER       { EASY_MACHINE_SET_LOCATION;
                   char *text = easy_machine_get_text (easy_machine_scanner);
                   int register_idx = strtol (text + 2, NULL, 10);
                   easy_machine_asm_append_register_parameter (parser_arg->routine,
                                                               register_idx); }
;

%%

/* FIXME: the error-reporting facility in generator/generator.y is derived
   from this, but improved. */
void
easy_machine_error (YYLTYPE *locp,
              struct parser_arg *parser_arg,
              yyscan_t easy_machine_scanner, const char *message)
{
  char *file_name = NULL;
  if (parser_arg != NULL)
    file_name = parser_arg->routine->file_name;
  if (file_name == NULL)
    file_name = "(input)";
  printf ("%s:%i: %s near \"%s\".\n",
          file_name,
          easy_machine_get_lineno (easy_machine_scanner), message,
          easy_machine_get_text (easy_machine_scanner));
  exit (EXIT_FAILURE);
}

__attribute__ ((noreturn)) static void
easy_machine_simple_error (void *easy_machine_scanner, const char *message)
{
  easy_machine_error (easy_machine_get_lloc (easy_machine_scanner),
                      NULL, /* We have no routine here, but it's not important. */
                      easy_machine_scanner,
                      message);
}

void
easy_machine_scan_error (void *easy_machine_scanner)
{
  easy_machine_simple_error (easy_machine_scanner, "scan error");
}

void
easy_machine_parse_error (void *easy_machine_scanner)
{
  easy_machine_simple_error (easy_machine_scanner, "parse error");
}

/* This is the main parser function doing all the work.  The other parsing
   functions, meant as part of a more convenient API for the user, all rely on
   this.
   The pointed scanner must be already initialized when this is called, and
   it's the caller's responsibility to finalize it. */
static void
easy_machine_parse_core (yyscan_t scanner, struct easy_machine_routine *p)
{
  struct parser_arg pa;
  pa.routine = p;
  /* FIXME: if I ever make parsing errors non-fatal, call easy_machine_lex_destroy before
     returning. */
  if (easy_machine_parse (& pa, scanner))
    easy_machine_error (easy_machine_get_lloc (scanner),
                        & pa,
                        scanner, "parse error");
}

void
easy_machine_parse_file_star (FILE *input_file, struct easy_machine_routine *p)
{
  yyscan_t scanner;
  easy_machine_lex_init (& scanner);
  easy_machine_set_in (input_file, scanner);

  easy_machine_parse_core (scanner, p);

  easy_machine_set_in (NULL, scanner);
  easy_machine_lex_destroy (scanner);
}

void
easy_machine_parse_file (const char *input_file_name, struct easy_machine_routine *p)
{
  FILE *f;
  if ((f = fopen (input_file_name, "r")) == NULL)
    {
      printf ("failed opening file");
      exit (EXIT_FAILURE);
    }

  p->file_name = easy_machine_string_clone (input_file_name);
  /* FIXME: if I ever make parse errors non-fatal, I'll need to close the file
     before returning. */
  easy_machine_parse_file_star (f, p);
  fclose (f);
}

void
easy_machine_parse_string (const char *string, struct easy_machine_routine *p)
{
  yyscan_t scanner;
  p->file_name = easy_machine_string_clone ("<literal-string>");
  easy_machine_lex_init (& scanner);
  easy_machine__scan_string (string, scanner);
  easy_machine_parse_core (scanner, p);
  easy_machine_lex_destroy (scanner);
}
