/* Easy Machine: assember header.
   This is a derivative work based on Jitter, which is released under
   the GNU General Public License, version 3 or later. */

#ifndef EASY_MACHINE_ASSEMBLER_H_
#define EASY_MACHINE_ASSEMBLER_H_

#include <config.h>

#include <easy-machine/jitter-dynamic-buffer.h>
#include <easy-machine/jitter-hash.h>

#include <easy-machine/easy-machine.h>
#include <easy-machine/easy-machine-machine.h>
#include <easy-machine/easy-machine-meta-instructions.h>




/* The rest.
 * ************************************************************************** */

enum easy_machine_asm_section
  {
    easy_machine_asm_section_code = 0,
    easy_machine_asm_section_readonly = 1,
    easy_machine_asm_section_readwrite = 2,

    easy_machine_asm_section_no = 3
  };

extern const char *
easy_machine_asm_section_names [easy_machine_asm_section_no];

extern easy_machine_word
easy_machine_section_load_addresses [easy_machine_asm_section_no];

enum easy_machine_asm_instruction_argument_kind
  {
    easy_machine_asm_instruction_argument_kind_numeric,
    easy_machine_asm_instruction_argument_kind_register,
    easy_machine_asm_instruction_argument_kind_label
  };

struct easy_machine_asm_instruction_argument
{
  enum easy_machine_asm_instruction_argument_kind kind;
  union
  {
    easy_machine_word numeric_literal;
    easy_machine_register_index register_idx;
    char *label_literal;
  };
};

struct easy_machine_asm_instruction
{
  /* The source line where the instruction began. */
  int line_no;

  char *mnemonic;

  /* This is determined at resolution time. */
  easy_machine_instruction_opcode opcode;

  size_t arg_no;
  struct easy_machine_asm_instruction_argument args [EASY_MACHINE_MAX_ARGS];
};

enum easy_machine_asm_element_kind
  {
    easy_machine_asm_element_kind_instruction,
    easy_machine_asm_element_kind_datum_byte,
    easy_machine_asm_element_kind_datum_word,
    easy_machine_asm_element_kind_datum_label,

    /* A named program point or datum location. */
    easy_machine_asm_element_kind_label
  };

struct easy_machine_asm_element
{
  /* The source line where the element began. */
  int line_no;
  enum easy_machine_asm_element_kind kind;

  /* Size in bytes, in non-symbolic form after everything is resolved. */
  size_t size;

  /* Offset from section beginning, in bytes, again in non-symbolic form after
     everything is resolved. */
  size_t offset_from_section;

  /* The kind-dependent field. */
  union
  {
    struct easy_machine_asm_instruction instruction;
    uint8_t byte;
    easy_machine_word word;
    char *label;
  };
};

struct easy_machine_asm_object_place
{
  /* Section id. */
  enum easy_machine_asm_section section;

  /* Offset from the beginning of the section. */
  easy_machine_word offset;
};

struct easy_machine_routine
{
  /* File name, if any, and source line number. */
  char *file_name;
  int line_no;

  /* The section we are currently assembling. */
  enum easy_machine_asm_section current_section;

  /* The instruction we are currently assembling, if any; NULL otherwise. */
  struct easy_machine_asm_instruction *current_instruction;
  bool appending_bytes;
  bool appending_words;

  /* For each section, keep a dynamic buffer.  Each buffer is a sequence of
     struct easy_machine_asm_element. */
  struct jitter_dynamic_buffer section_contents [easy_machine_asm_section_no];

  /* The size of each section, in bytes. */
  size_t section_sizes [easy_machine_asm_section_no];

  /* The size of each section, in bytes, rounded up to a whole number of
     pages. */
  size_t section_sizes_rounded_up [easy_machine_asm_section_no];

  /* A string hash mapping each label name to a virtual address. */
  struct jitter_hash_table label_to_place;
};

void
easy_machine_asm_initialize_routine (struct easy_machine_routine *r);

void
easy_machine_asm_enter_section (struct easy_machine_routine *r,
                                enum easy_machine_asm_section section);

void
easy_machine_asm_append_words (struct easy_machine_routine *r);
void
easy_machine_asm_append_bytes (struct easy_machine_routine *r);

void
easy_machine_asm_append_datum_symbolic_label (struct easy_machine_routine *r,
                                              const char *name);
void
easy_machine_asm_append_datum_literal (struct easy_machine_routine *r,
                                       easy_machine_word w);

void
easy_machine_asm_append_symbolic_label (struct easy_machine_routine *r,
                                        const char *name);


void
easy_machine_asm_append_symbolic_label (struct easy_machine_routine *r,
                                        const char *name);

void
easy_machine_asm_append_instruction_name (struct easy_machine_routine *r,
                                          const char *name);

void
easy_machine_asm_append_literal_parameter (struct easy_machine_routine *r,
                                           easy_machine_word w);

void
easy_machine_asm_append_register_parameter (struct easy_machine_routine *r,
                                            easy_machine_register_index idx);

void
easy_machine_asm_append_symbolic_label_parameter (struct easy_machine_routine *r,
                                                  const char *name);

void
easy_machine_asm_load (struct easy_machine_routine *r,
                       struct easy_machine_assmbler_command_line *cl);

#endif // #ifndef EASY_MACHINE_ASSEMBLER_H_
