/* Easy Machine: Code generator driver.
   This is a derivative work based on Jitter, which is released under
   the GNU General Public License, version 3 or later. */

/* Use Gnulib. */
# include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdbool.h>

#include <argp.h>

#include <easy-machine/easy-machine-generate-code.h>


/* Command line handling.
 * ************************************************************************** */

/* Update our option state with the information from a single command-line
   option. */
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct command_line *cl = state->input;
  switch (key)
    {
      /* FIXME: within the case I could use state->arg_num to get the current
         non-option argument index, if needed. */
    case ARGP_KEY_INIT:
      /* Set reasonable default values. */
      cl->debug = false;
      break;
    case 'd':
      cl->debug = true;
      break;

    case ARGP_KEY_ARG:
      //cl->input_file = arg;
      break;

    /* case ARGP_KEY_NO_ARGS: /\* If this case is omitted, the default is sensible. *\/ */
    /*   argp_error (state, "you have to supply the input file\n"); */
    /*   break; */

    /* case ARGP_KEY_SUCCESS: /\* If this case is omitted, the default is sensible. *\/ */
    /*   break; */

    case ARGP_KEY_END:
      if (state->arg_num != 0)
        argp_error (state, "you gave %i input files instead of zero.",
                    (int)state->arg_num);
      break;

    /* case ARGP_KEY_FINI: /\* If this case is omitted, the default is sensible. *\/ */
    /*   break; */

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

/* Command-line option specification. */
static struct argp_option main_option_specification[] =
  {/* Most frequently used options. */
   {NULL, '\0', NULL, OPTION_DOC, "Frequently used options:", 10},
   {"debug", 'd', NULL, 0,
    "Print out debugging information"},

   /* Common GNU-style options. */
   {NULL, '\0', NULL, OPTION_DOC, "Common GNU-style options:", -1},
   /* These are automatically generated. */

   /* Option terminator. */
   { 0 }};

const char *argp_program_version
  = "C code generator (" EASY_MACHINE_PACKAGE_NAME ") "
    EASY_MACHINE_PACKAGE_VERSION "\n"
    "Copyright (C) 2016, 2017, 2019 Luca Saiu.\n"
    "This is a derivative work based on Jitter, which is released under\n"
    "the GNU General Public License, version 3 or later.\n"
    "Jitter comes with ABSOLUTELY NO WARRANTY.\n"
    "You may redistribute copies of Jitter under the terms of the\n"
    "GNU General Public License, version 3 or any later version published by\n"
    "the Free Software Foundation.  For more information see the file named\n"
    "COPYING in the source distribution or the git repository.\n\n"
    "Written by Luca Saiu <http://ageinghacker.net> (Jitter, its runtime, "
    "this derivative program).";
const char *argp_program_bug_address = EASY_MACHINE_PACKAGE_BUGREPORT;

/* The parser main data structure. */
static struct argp argp =
  {
    main_option_specification,
    parse_opt,
    NULL,
    "Generate C code from the hand-written instruction specification."
  };




/* Main.
 * ************************************************************************** */

int
main (int argc, char **argv)
{
  /* Parse our arguments; every option seen by 'parse_opt' will
     be reflected in 'arguments'. */
  struct command_line cl;
  argp_parse (&argp, argc, argv,
              0,//ARGP_IN_ORDER,
              0, & cl);

  /* Do the actual work. */
  easy_machine_generate_code_do_work (& cl);
  return EXIT_SUCCESS;
}
