#include <config.h>

#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <unistd.h>

#include <easy-machine/easy-machine-utility.h>
#include <easy-machine/easy-machine-machine.h>
#include <easy-machine/easy-machine-meta-instructions.h>
#include <easy-machine/easy-machine-instructions-machine-generated.h>

static void
easy_machine_syscall (struct easy_machine_thread_state *s,
                      easy_machine_word syscall_id,
                      easy_machine_word arg1,
                      easy_machine_word arg2,
                      struct easy_machine_assmbler_command_line *cl)
{
  switch (syscall_id)
    {
    case 0:
      printf ("%c", arg1);
      fflush (stdout);
      break;

    case 1:
      s->registers [0] = getchar ();
      break;

    case 2:
      if (cl->debug)
        printf ("Exiting with success\n");
      else
        printf ("\n");
      exit (EXIT_SUCCESS);
      break;

    case 3:
      sleep (arg1);
      break;

    default:
      easy_machine_exception ("invalid syscall 0x%lx",
                              (unsigned long) syscall_id);
    }
}

/* Include the machine-generated C file with the main instruction interpreter
   definition. */
#include "easy-machine-machine-interpreter-machine-generated.c"

extern easy_machine_word
easy_machine_virtual_memory_beginning;
extern easy_machine_word
easy_machine_virtual_memory_past_end;

void
easy_machine_run (struct easy_machine_thread_state *s,
                  struct easy_machine_assmbler_command_line *cl)
{
  if (cl->debug)
    printf ("virtual addresses: [0x%lx, 0x%lx)\n",
            (unsigned long) easy_machine_virtual_memory_beginning,
            (unsigned long) easy_machine_virtual_memory_past_end);
  do
    {
      s->vpc = easy_machine_interpret_instruction (s, cl);
      fflush (stdout);
    }
  while (true);
}
