/* Easy Machine.

   Machine interpreter: non-generated code. */


#include <config.h>

#include <stdio.h>
#include <string.h>

#include <xalloc.h>

#include <easy-machine/easy-machine-machine.h>
#include <easy-machine/easy-machine-utility.h>
#include <easy-machine/easy-machine-meta-instructions.h>
#include <easy-machine/easy-machine-machine-interpreter.h>


/* Command line.
 * ************************************************************************** */

void
easy_machine_machine_command_line_initialize
   (struct easy_machine_machine_command_line *cl)
{
  cl->debug = false;
}




/* Machine global state.
 * ************************************************************************** */

/* FIXME: this is a stub, just to be used before I implement paging. */
const size_t
easy_machine_memory_size_in_megabytes = 64;//1024;
const size_t easy_machine_memory_size_in_bytes
  = easy_machine_memory_size_in_megabytes * 1024 * 1024;
const size_t easy_machine_memory_size_in_words
  = easy_machine_memory_size_in_bytes / sizeof (easy_machine_word);
static uint8_t
*easy_machine_memory_content = NULL;

easy_machine_word
easy_machine_virtual_memory_beginning = 0;
easy_machine_word
easy_machine_virtual_memory_past_end = 0;

void
easy_machine_initialize_global_state (void)
{
  easy_machine_memory_content
    = xmalloc (easy_machine_memory_size_in_bytes);
  memset (easy_machine_memory_content, 0xff, easy_machine_memory_size_in_bytes);
  easy_machine_virtual_memory_beginning
    = EASY_MACHINE_MINIMUM_ADDRESS;
  easy_machine_virtual_memory_past_end
    = (easy_machine_virtual_memory_beginning
       + easy_machine_memory_size_in_bytes);
}




/* Memory access.
 * ************************************************************************** */

uint8_t *
easy_machine_virtual_address_to_pointer (easy_machine_word address)
{
  if (address < easy_machine_virtual_memory_beginning)
    easy_machine_fatal ("invalid virtual address 0x%lx", (unsigned long) address);
  else if (address >= easy_machine_virtual_memory_past_end)
    easy_machine_fatal ("invalid virtual address 0x%lx", (unsigned long) address);

  return (easy_machine_memory_content
          + address - easy_machine_virtual_memory_beginning);
}

easy_machine_word
easy_machine_pointer_to_virtual_address (uint8_t *p)
{
  if (p < easy_machine_memory_content)
    easy_machine_fatal ("out-of-bound C array address %p", p);
  else if (p >= (easy_machine_memory_content
                 + easy_machine_memory_size_in_megabytes))
    easy_machine_fatal ("out-of-bound C array address %p", p);

  return ((p - easy_machine_memory_content)
          + easy_machine_virtual_memory_beginning);
}

uint8_t
easy_machine_load_byte (easy_machine_word address)
{
  uint8_t *p = easy_machine_virtual_address_to_pointer (address);
  return * p;
}

void
easy_machine_store_byte (easy_machine_word address,
                         uint8_t datum)
{
  uint8_t *p = easy_machine_virtual_address_to_pointer (address);
  * p = datum;
}

union easy_machine_word_or_bytes
{
  easy_machine_word word;
  uint8_t bytes [sizeof (easy_machine_word)];
};

easy_machine_word
easy_machine_load_word (easy_machine_word address)
{
  const size_t byte_no = sizeof (easy_machine_word);
  int i;
  // FIXME: I can enforce my preference of endianness here.
  union easy_machine_word_or_bytes copy;
  for (i = 0; i < byte_no; i ++)
    copy.bytes [i] = easy_machine_load_byte (address + i);
  return copy.word;
}

void
easy_machine_store_word (easy_machine_word address,
                         easy_machine_word datum)
{
  const size_t byte_no = sizeof (easy_machine_word);
  // FIXME: I can enforce my preference of endianness here.
  union easy_machine_word_or_bytes copy;
  copy.word = datum;
  int i;
  for (i = 0; i < byte_no; i ++)
    easy_machine_store_byte (address + i, copy.bytes [i]);
}




/* Machine thread state.
 * ************************************************************************** */

void
easy_machine_initialize_thread_state (struct easy_machine_thread_state *s,
                                      easy_machine_word vpc)
{
  int i;
  for (i = 0; i < EASY_MACHINE_REGISTER_NO; i ++)
    s->registers [i] = -1;
  s->vpc = vpc;
}
