/* Easy Machine.

   Instruction definitions: header. */

#ifndef EASY_MACHINE_META_INSTRUCTIONS_H_
#define EASY_MACHINE_META_INSTRUCTIONS_H_

#include <stdlib.h>
#include <stdint.h>

#include <easy-machine/easy-machine.h>


/* Hand-written instruction specifications.
 * ************************************************************************** */

/* An instruction argument kind.  These can be or-ed together into a bitmask
   when specifying kinds for a meta-instruction.  Every value is guaranteed to
   be a power of two. */
enum easy_machine_instruction_argument_kind
  {
    easy_machine_instruction_argument_kind_immediate = 1,
    easy_machine_instruction_argument_kind_register  = 2,

    /* This is not a valid argument kind: it is used as a limit when iterating
       over the values of this enum. */
    easy_machine_instruction_argument_kind_terminator = 4
  };

/* How many arguments an instruction can have, at most. */
#define EASY_MACHINE_MAX_ARGS 3

/* An instruction or meta-instruction descriptor. */
struct easy_machine_instruction_descriptor
{
  /* The instruction mnemonic, used for assembling and disassembling. */
  char *mnemonic;

  /* A human-readable synopsis of what the instruction does. */
  char *synopsis;

  /* A human-readable description of what the instruction does. */
  char *description;

  /* How many arguments the instruction takes. */
  size_t arg_no;

  /* Argument specification.  Only the first arg_no elements are used.  If this
     struct specifies a non-meta instruction then each used element must have
     exactly one bit on: no alternatives are allowed. */
  enum easy_machine_instruction_argument_kind args [EASY_MACHINE_MAX_ARGS];

  /* Implementation in C. */
  char *source;
};

/* An array containing a descriptor for every meta-instruction. */
extern const struct easy_machine_instruction_descriptor
easy_machine_meta_instruction_descriptors [];

/* How many meta-instruction descriptors there are. */
extern const size_t
easy_machine_meta_instruction_descriptor_no;




/* Instruction encoding.
 * ************************************************************************** */

typedef int
easy_machine_instruction_opcode;

struct easy_machine_instruction
{
  int32_t opcode;
  int32_t arg0;
  int32_t arg1;
  int32_t arg2;
} __attribute__ ((packed));

/* Defined in easy-machine-machine-instruction-names-machine-generated.c . */
extern const char
*easy_machine_instruction_names [];




/* System parameters.
 * ************************************************************************** */

#define EASY_MACHINE_PAGE_SIZE       (1024 * 4)
#define EASY_MACHINE_MINIMUM_ADDRESS (1024 * 1024)

#endif // #ifndef EASY_MACHINE_META_INSTRUCTIONS_H_
